/*
 * modbus-tcp - context.c
 *	version 0.1.1
 *
 *	Context
 *
 * Copyright (C) 2007..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <modbus.h>
#include <compiler.h>
#include <context.h>
#include <list.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <rejd_byteorder.h>

#if CONFIG_DEBUG
#define debug(fmt...) fprintf(stderr, fmt)
#else
#define debug(fmt...) do { } while (0)
#endif

static LIST_HEAD(orphan_queue);

extern int conn_count;

int context_init(struct c_context *c, int fd)
{
	c->fd = fd;
	buffer_init(&c->receive_hdr);
	buffer_init(&c->receive);
	buffer_skip(&c->receive, 4);
	c->send = NULL;
	INIT_LIST_HEAD(&c->req_ready);
	INIT_LIST_HEAD(&c->req_wait);
	c->rcount = 0;
	c->write_wait = 0;
	return 0;
}

int context_destroy(struct c_context *c)
{
	struct list_head *t;
	struct c_request *r;

	BUG_ON(!conn_count);
	while (!list_empty(&c->req_wait)) {
		t = list_next(&c->req_wait);
		list_del(t);
		r = list_entry(t, struct c_request, list);

		r->ctx = NULL;
		/*
		 * current kernel interface doesn't support cancel
		 */
		list_add(t, &orphan_queue);
	}

	while (!list_empty(&c->req_ready)) {
		t = list_next(&c->req_ready);
		r = list_entry(t, struct c_request, list);
		request_destroy(r);
		list_del(t);
	}
	list_del(&c->list);
	if (verbose_level > 0)
		fprintf(stderr, "closing connection: %d\n", c->fd);
	close(c->fd);
	free(c);
	conn_count--;
	return 0;
}

int context_write_schedule(struct c_context *c)
{
	/* already scheduled */
	if (c->send != NULL)
		return 0;

	/* nothing to schedule */
	if (list_empty(&c->req_ready))
		return -ENOENT;

	/* pick first completed request */
	c->send = list_entry(c->req_ready.next, struct c_request, list);
	list_del(&c->send->list);
	return 0;
}

int context_ready(struct c_context *c)
{
	context_write_schedule(c);
	return (c->send != NULL);
}

struct modbus_error_msg {
	uint8_t cmd;
	uint8_t code;
} __attribute__((packed));

int generate_modbus_error(struct c_request *r, unsigned int code)
{
	struct mbap_header *hdr;
	struct modbus_error_msg *err;
	unsigned int rlen = sizeof(*hdr) + sizeof(*err);
	void *d = malloc(rlen);

	if (d == NULL)
		return -ENOMEM;
	r->mreq.r = d;
	r->mreq.r_len = rlen;
	hdr = d;

	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(r->mreq.r_len) - 6;
	hdr->seq = cpu_to_be16(r->tcp_id);
	hdr->addr = r->tcp_addr;

	err = (void *)(hdr + 1);
	err->cmd = ((uint8_t *) r->mreq.c)[4] | 0x80;
	err->code = code;
	return 0;
}

int context_recv_mbap(struct c_context *c, struct mbap_header *hdr,
		void *buf, unsigned int len)
{
	struct c_request *r;
	int ret;

	r = malloc(sizeof(*r));
	if (r == NULL)
		return -ENOMEM;

	r->tcp_id = be16_to_cpu(hdr->seq);
	r->tcp_addr = hdr->addr;
	r->ctx = c;
	r->sent = 0;

	ret = modbus_request_init(&r->mreq);
	if (ret) {
		free(r);
		goto out;
	}
	r->mreq.c = buf;
	r->mreq.c_len = len;
	modbus_request_fill_id(&r->mreq);

	debug("MBAP: received packet id=%d, kernel_id=%d, len=%d\n",
			r->tcp_id, r->mreq.id, len - 4);
	modbus_set_device(&modbus, hdr->addr);
	ret = modbus_send(&modbus, &r->mreq);
	if (ret) {
		generate_modbus_error(r, MBEX_GWERR);
		list_add(&r->list, &c->req_ready);
	} else {
		list_add(&r->list, &c->req_wait);
	}
	c->rcount++;
out:
	return ret;
}

int context_read(struct c_context *c)
{
	struct mbap_header *hdr;
	int ret;
	unsigned int len;

	if (c->receive_hdr.length < sizeof(*hdr)) {
		ret = buffer_read(c->fd, &c->receive_hdr, sizeof(*hdr));
		if (ret < 0) {
			context_destroy(c);
			return -ECONNRESET;
		}
		if (!ret)
			c->write_wait = 1;
		return 0;
	}

	hdr = c->receive_hdr.data;
	len = be16_to_cpu(hdr->len);
	if ((len < sizeof(*hdr) - 6) || (len > MBAP_MTU - 6)) {
		context_destroy(c);
		return -ECONNRESET;
	}
	len += 6;

	ret = buffer_read(c->fd, &c->receive, len + 4 - sizeof(*hdr)
			- c->receive.length);
	if (ret <= 0) {
		if (!ret) {
			c->write_wait = 1;
			return 0;
		} else {
			context_destroy(c);
			return -ECONNRESET;
		}
	}

	if (c->receive.length < len + 4 - sizeof(*hdr))
		return 0;
	context_recv_mbap(c, hdr, c->receive.data, c->receive.length);

	buffer_destroy(&c->receive_hdr);
	buffer_init(&c->receive_hdr);
	buffer_init(&c->receive);
	buffer_skip(&c->receive, 4);
	return 0;
}

/*
 * send response to client
 */
int context_write(struct c_context *c)
{
	struct c_request *r;
	int ret = 0;
	int to_write;

	if (c->send == NULL)
		context_write_schedule(c);
	BUG_ON(c->send == NULL);
	r = c->send;

	to_write = r->mreq.r_len - r->sent;
	ret = write(c->fd, (char *)r->mreq.r + r->sent, to_write);
	if (ret < 0)
		return -ret;
	if (ret > 0)
		r->sent += ret;
	if (ret == to_write) {
		c->rcount--;
		request_destroy(r);
		c->send = NULL;
		context_write_schedule(c);
	}
	return ret;
}

void c_req_add(struct c_context *c, struct c_request *r)
{
#if 0
	BUG_ON(c == NULL);
	cr->r = r;
	cr->p = p;
	cr->ops = ops;
	cr->ctx = c;
	cr->written = 0;
	list_add_tail(&cr->list, &c->requests);
	c->rcount++;
#endif
}

void request_destroy(struct c_request *r)
{
	modbus_request_destroy(&r->mreq);
	free(r);
}

int c_req_completed(struct c_request *cr)
{
	return cr->mreq.r_len == cr->sent;
}

void orphan_killer(void)
{
	/* TODO: do sanity checks for timedout requests */
}
