/*
 * ar2c-lp-proxy - mclock.c
 * 	version 0.1.2
 *
 * 	Monotonic clock
 * 
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <mclock.h>
#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HAVE_GETTIMEOFDAY
# include <sys/time.h>
#endif
#ifdef HAVE_COMPILER_H
# include <compiler.h>
#endif

#ifndef likely
# define likely(x)	(x)
#endif

#ifndef unlikely
# define unlikely(x)	(x)
#endif

/*
 * POSIX monotonic clock - preferred clock
 *
 * POSIX 1003.1-2001
 */
#if defined(_POSIX_MONOTONIC_CLOCK)
# define monotonic_clock_get(ts) clock_gettime(CLOCK_MONOTONIC, ts)
# if _POSIX_MONOTONIC_CLOCK == 0
static int __have_monotonic_clock;
#  define have_monotonic_clock() __have_monotonic_clock

#  define HAVE_MONOTONIC_INIT
static void monotonic_init(void)
{
	struct timespec ts;

	__have_monotonic_clock = sysconf(_SC_MONOTONIC_CLOCK);
	if (__have_monotonic_clock < 0)
		if (monotonic_clock_get(&ts) != -1)
			__have_monotonic_clock = 1;
}
# else
#  define have_monotonic_clock() _POSIX_MONOTONIC_CLOCK
# endif
#else
# define have_monotonic_clock() (-1)
# define monotonic_clock_get(ts) (-EINVAL)
#endif

#ifndef HAVE_MONOTONIC_INIT
static void monotonic_init(void)
{
}
#endif

/*
 * POSIX realtime clock - high resolution but not monotonic
 *
 * POSIX 1003.1-2001
 */
#if defined(_POSIX_TIMERS)
# define realtime_clock_get(ts) clock_gettime(CLOCK_REALTIME, ts)
# if _POSIX_TIMERS == 0
static int __have_realtime_clock;
#  define have_realtime_clock() __have_realtime_clock

#  define HAVE_REALTIME_INIT
static void realtime_init(void)
{
	__have_realtime_clock = sysconf(_SC_POSIX_TIMERS);
}
# else
#  define have_realtime_clock() _POSIX_TIMERS
# endif
#else
# define have_realtime_clock() (-1)
# define realtime_clock_get(ts) (-EINVAL)
#endif

#ifndef HAVE_REALTIME_INIT
static void realtime_init(void)
{
}
#endif

#ifdef HAVE_TIME
# define have_time()	1
#else
# define have_time()	-1
#endif

/*
 * realtime_fixup - monotonize realtime 
 */
static void realtime_fixup(struct timespec *new)
{
	static struct timespec old, delta;

	/*
	 * recalculate delta when new < old
	 */
	if (unlikely(new->tv_sec < old.tv_sec
				|| (new->tv_sec == old.tv_sec
					&& new->tv_nsec < old.tv_nsec))) {
		/* delta = old - new */
		delta.tv_sec = old.tv_sec - new->tv_sec;
		if (old.tv_nsec < new->tv_nsec) {
			delta.tv_sec--;
			delta.tv_nsec += 1000000000UL;
		}
		delta.tv_nsec = old.tv_nsec - new->tv_nsec;
	}

	/* new += delta */
	new->tv_sec = new->tv_sec + delta.tv_sec;
	new->tv_nsec = new->tv_nsec + delta.tv_nsec;
	if (new->tv_nsec >= 1000000000L) {
		new->tv_nsec -= 1000000000L;
		new->tv_sec++;
	}
	old = *new;
}

/*
 * my_clock_gettime - high resolution monotonic clock
 */
static int my_clock_gettime(struct timespec *ts)
{
	int ret;

	if (have_monotonic_clock() > 0)
		return monotonic_clock_get(ts);

	if (have_realtime_clock() > 0) {
		ret = realtime_clock_get(ts);
		if (ret)
			return ret;
		realtime_fixup(ts);
	}
#ifdef HAVE_GETTIMEOFDAY
	{
		struct timeval tv;

		ret = gettimeofday(&tv, NULL);
		if (ret)
			return ret;
		ts->tv_sec = tv.tv_sec;
		ts->tv_nsec = 1000 * tv.tv_usec;
		realtime_fixup(ts);
		return 0;
	}
#endif
#ifdef HAVE_TIME
	{
		ts->tv_sec = time(NULL);
		ts->tv_nsec = 0;
		realtime_fixup(ts);
		return 0;
	}
#endif
	return -ENOSYS;
}

static void clock_not_working(void)
{
	perror("mclock: clock error: %s\n");
	exit(EXIT_FAILURE);
}

/*
 * time_ns - get number of nanoseconds of mclock
 */
unsigned long long time_ns(void)
{
	struct timespec ts;
	int ret;

	ret = my_clock_gettime(&ts);
	if (ret)
		clock_not_working();
	return ts.tv_sec * 1000000000ULL + ts.tv_nsec;
}

/*
 * time_ms - get number of miliseconds of mclock
 */
unsigned long long time_ms(void)
{
	struct timespec ts;
	int ret;

	ret = my_clock_gettime(&ts);
	if (ret)
		clock_not_working();
	return ts.tv_sec * 1000ULL + ts.tv_nsec / 1000000;
}

/*
 * time_s - get number of seconds of mclock
 */
time_t time_s(void)
{
	struct timespec ts;
	int ret;

	ret = my_clock_gettime(&ts);
	if (ret)
		clock_not_working();
	return ts.tv_sec;
}

/*
 * mclock_init - initialize mclock
 * 
 * This function returns zero on success; error code otherwise.
 */
int mclock_init(void)
{
	struct timespec ts;

	monotonic_init();
	realtime_init();

	/* just return if clock works */
	return my_clock_gettime(&ts);
}
